import cgs
import numpy as npy
import scipy.optimize as sciop
import scipy.integrate as sciint

def nnucl_spont (rhog, pgas, mseed, sigdotN, past, sigast, gz):
    """
    nucleation rate profile
    """

    dsig_p =  sigdotN *1/(sigast*pgas*npy.sqrt(2*npy.pi)) \
                *npy.exp(-0.5 *(npy.log(pgas/past)/sigast)**2)

    rhodotN = dsig_p *rhog*gz
    return rhodotN


def v_settl (tstop, gz):
    return tstop*gz


def mass_to_size (mass, rhoint):
    """
    returns the total size, taking accound condensate mass 
    ("mass") and seed mass ("mseed")
    """
    size = npy.cbrt(0.75*mass/(npy.pi*rhoint))
    return size


def get_pprop (xseed, xc, Tgas, rhog, mseed, 
                rhointP, gz, mgas=2.34*cgs.amu, EpsteinApprox=False):
    """
    get the physical properties of the particles:
        number density (np), mass (mp), radius (ap), and aerodynamics (tstop, vsettl)
    """

    #particle number and condensate mass density
    np =   npy.maximum(xseed*rhog/mseed, 0)
    rhoc = npy.maximum(rhog*xc, 0)
    ii = (rhoc>0)*(np>0) 

    mc = npy.zeros_like(rhoc)
    mc[ii] = rhoc[ii]/np[ii]
    mp = mc +mseed

    #cloud particle size
    ap = mass_to_size (mp, rhointP)

    if EpsteinApprox:
        tstop = ap*rhointP/(v_th(Tgas,mgas)*rhog)
        vsettl = v_settl(tstop, gz)
    else:
        import myphys.aero as aero
        tstop, vsettl = aero.t_stopT(ap,rhog,Tgas,v_settl,vgasArgs=(gz,),
                                     mgas=mgas,rhoint=rhointP)
    return np, mp, ap, tstop, vsettl


def es_X (T, species, units='bar', **kwargs):
    """
    general formula for saturation pressure of species X
    """
    if species=='NH3':
        es = es_NH3 (T, units, **kwargs)
    elif species=='MgSiO3':
        es = es_MgSiO3 (T, units, **kwargs)
    elif species=='KCl':
        es = es_KCl (T, units, **kwargs)

    if units!='bar':
        es *= cgs.units_conv('pressure', 'bar', units)

    return es


def rhoEq_X (T, species, mvap, units='cgs', **kwargs):
    """
    general formula for equilibrium density species X
    """

    esX = es_X (T, species, units='cgs', **kwargs)
    rhoEq = mvap *esX/(cgs.kB*T)

    return rhoEq


def v_th (T, mass):
    """
    mean speed
    """
    vth = npy.sqrt(8*cgs.kB*T/(npy.pi*mass))
    return vth


def inv_tcond (ap, np, vth, diffM):
    """
    inverse condensation timescale

    TBD: fix factor 4
    """
    rate = npy.minimum(npy.pi*vth*ap, 4*npy.pi*diffM) *ap*np
    return rate


def inv_tcoag (ap, np, vsed, vBM, Dp):
    """
    coagulation rate
    - 4 is the combined cross section
    - 0.5 for double counting
    - 0.5 for differential correction of similar particles
    """
    invtcoag = 0.5*(4*npy.pi *0.5*vsed*ap        *ap*np +\
                    4*npy.pi *npy.minimum(vBM*ap,Dp) *ap*np)

    return invtcoag


def rho_eq_MgSiO3 (T, units='cgs', mvap=100*cgs.amu):
    """
    returns equilibrium density (cgs units)
    """

    pres = es_MgSiO3 (T, units='cgs')
    rhoEq = mvap*pres /(cgs.kB*T)

    return rhoEq


def es_MgSiO3 (T, units='bar'):
    """
    return the pressure based on above formula
    """

    #following Ackerman & Marley (2001), Eq. A3
    pres = npy.exp(25.37 -58663/T)
    return pres


def es_KCl (T, units='bar'):
    """
    Morley et al. (2012) [in bar]
    """
    pres = 10**(7.611 -11382./T)
    #pres = 10**(14.2 -11376./T)

    return pres


def X_eq (T, rhog, species, mvap):
    """
    the equilibrium concentration
    """

    rhoeq = rhoEq_X (T, species, mvap, units='cgs')
    Xeq = rhoeq /rhog

    return Xeq


def diff_i (pgas, T, mgas=2.34*cgs.amu, sigmolC=8e-15):
    """
    It's just ~1/3 v_th l_mfp but a bit more complicated b/c
    of mixing

    We follow Woitke & Helling (more or less)
    """
    vth_red = v_th(T, mgas)
    diff = cgs.kB*T/(3*sigmolC*pgas) *vth_red

    return diff


def diff_p (rp, rhog, T, mgas, sigmol):
    """
    Particle diffusivity as in Stokes Einstein relation
    """
    lmfp = (npy.sqrt(2.) *rhog/mgas *sigmol)**-1
    vth = v_th (T, mgas)
    eta = rhog *0.5*vth*lmfp #Chapman

    #Stokes-Einstein formula
    Di = cgs.kB*T /(6*npy.pi*eta*rp)
    return Di


def calc_tau_vert (zarr, kappa):
    """
    """
    out = sciint.simps (kappa, zarr)
    return out


def calc_tau_limb (zarr, kappa, Rplanet):
    """
    this assumes theta is small
    factor 2 accounts for negative theta integration!
    """
    tharr = npy.sqrt(2*(zarr-zarr.min())/Rplanet)
    out = 2*sciint.simps (kappa*Rplanet, tharr)

    return out
