import numpy as npy
import sys
sys.path.append('./../Profiles/')
sys.path.append('./../Genfuns/')
import genfuns3 as gf
import planprofile as planprof
import cgs
import scipy.integrate as sciint
from scipy.special import erf
import config as cfg
import copy

def bvp_negative (bvpsol, iphase):
    """
    consider a solution invalid when it contains a 
    negative condensate density (xc -- 3rd index)
    """
    con3 = True
    if iphase<=1:
        if any(bvpsol.y[3][1:-1]/bvpsol.y[3].max()<=-1e-12):
            con3 = False
    else:
        if bvpsol.y[3].min()/bvpsol.y[3].max()<=-1e-12:
            con3 = False

    if any(npy.isnan(bvpsol.y[3])):
        con3 = False
    return con3


def prepare_Ynew (Yg):
    """
    this removes some negative entries
    (TBR?)
    """
    Yn = copy.copy(Yg)
    ii = Yn[3,:]<0
    Yn[3,ii] = 0.
    return Yn


def get_from_bvpsol (xarr, bvpsol):
    zarr = x_to_z(xarr)
    lparr = cfg.lp_z (zarr)
    parr = npy.exp(lparr)

    Y = bvpsol.sol(xarr)
    Mn, Mc, xn, xc, xv = (Y.T *cfg.normArr).T

    #other quantities
    rhogarr = cfg.rho_z (zarr)

    Tarr = cfg.T_z (zarr)
    xeq = gf.X_eq (Tarr, rhogarr, cpars.species, cpars.mvap)
    return Mc,xc,xv,Mn,xn,rhogarr,Tarr,xeq,parr


def boundary_conditions (YB, YT):
    """
    0       :mass flux seeds
    1       :mass flux condensates
    2       :concentration seeds
    3       :concentration condensates
    4/-1    :concentration vapor

    YB: bottom BC
    YT: top BC
    """
    #top mass flux of nuclei
    lT = cfg.lp_z (cfg.ztop) #TBD check w/r this canges in phase 3
    pT = npy.exp(lT)
    #fraction of the flux at the top
    fT = get_fT (pT)
    y0T = fT *cpars.sigdotN /cfg.normArr[0]


    #0 mass flux at top
    BcTopL = [YT[0]+y0T, YT[1]]
    BcBotL = [YB[2], YB[3], YB[-1]-1.]

    return npy.array(BcBotL+BcTopL)


def get_fT (pT):
    """
    The fraction of the flux above the computational domain!
    """
    fT = 0.5 +0.5*erf(npy.log(pT/cpars.past) /(npy.sqrt(2.)*cpars.sigast))
    return fT


def initialize (pextent, ddir, doPlot):
    global cpars
    sys.path.append(ddir)
    import cloudpars as cpars


    fname = 'profile.txt'
    pbot = planprof.plot_profile(ddir+fname, cpars.Prange, cpars.species, 
                                    cpars.mvap, Xn=cpars.xvbot, doPlot=doPlot)
    pbot *= 1e6 #to cgs units

    #not the Prange in the table/cpars!
    Prange = npy.array([pbot, pbot/pextent])

    cfg.lp_z, cfg.z_lp, cfg.T_z, cfg.rho_z =\
            planprof.get_profile(fname=ddir+fname,
                        dunits={'pressure':'cgs', 'height':'cm', 'density':'cgs'},
                        order='cubic')

    cfg.zrange = cfg.z_lp(npy.log(Prange))
    cfg.Lfac = cfg.zrange[1] -cfg.zrange[0]
    cfg.zbot = cfg.zrange[0]
    cfg.ztop = cfg.zrange[1]

    #max.possible entry (from Table)
    zmax = cfg.z_lp(npy.log(cpars.Prange[1]*1e6))
    cfg.xmax = (zmax-cfg.zbot)/cfg.Lfac

    cfg.Ndim = 5
    Ns = 11
    xg = npy.linspace(0, 1., Ns)
    guess = (0., 0., 1., 0., 1.)
    Yg = (npy.array(guess) *npy.ones((Ns,cfg.Ndim))).T
    zg = npy.linspace(cfg.zbot, cfg.ztop, Ns)

    #normalizations (important)
    #initial normalization (normArr) is arbitrary
    #normalizations will be adjusted such that the numerical
    #values (Y) are ~unity
    #inorm indicates how to adjust:
    #-1 :adjust; quantity is negative (mass fluxes)
    #0  :do not adjust
    #1  :adjust for positive quantity (concentrations)
    cfg.normArr = npy.array([1e-10, 1e-10, 1e-10, 1.,  cpars.xvbot])
    cfg.inorm =            [-1,    -1,    1,     1,     0.]

    return xg, Yg


def x_to_z (x):
    return cfg.zbot +cfg.Lfac*x


def Y_x (x, Y):
    z = x_to_z(x)
    return cfg.Lfac*Y_z (z, Y)


def Y_z (z, Y):
    #convert to physical quantities
    #Mn :mass flux nuclei
    #Mc :condensate mass flux
    #xn :nuclei mass concentration
    #xc :condensate mass concentration
    #xv :vapor mass concentration
    Mn, Mc, xn, xc, xv = (Y.T *cfg.normArr).T

    #physical structure
    lp =   cfg.lp_z(z)
    pgas = npy.exp(lp)
    rhog = cfg.rho_z(z)
    Tgas = cfg.T_z(z)

    #get the equilibrium density
    rhoeq = gf.rhoEq_X(Tgas, cpars.species, cpars.mvap)
    vthZ =  gf.v_th(Tgas, cpars.mvap)

    #get the particle properties
    #mass size, and stopping times etc
    np, mp, ap, tstop, vsettl =\
            gf.get_pprop (xn, xc, Tgas, rhog, cpars.mseed, cpars.rhoint, cpars.gz,
                            mgas=cpars.mgas, EpsteinApprox=cpars.EpsteinApprox)

    #coagulation rate (n*sigma*delv)*n
    #0.5 from Krijt
    #counting correction (1/2)
    if cpars.doCoag:
        #first term is due to differential settling
        #second term due to diffusion
        vBM = gf.v_th(Tgas, 0.5*mp)
        Dp =  gf.diff_p(ap, rhog, Tgas, cpars.mgas, cpars.sigmol)

        coagrate = gf.inv_tcoag (ap, np, vsettl, vBM, Dp)
        dndt_coag = np *coagrate *cfg.ffac
    else:
        dndt_coag = 0.

    #mean free path vapor particles
    diffM = gf.diff_i(pgas, Tgas, cpars.mgas, cpars.sigmolC)
    rate = gf.inv_tcond(ap, np, vthZ, diffM)

    #condensation and nucleation rates
    feq = 1.
    rhodotC = rate *cpars.fstick *(xv*rhog -rhoeq*feq) *cfg.ffac
    rhodotN = gf.nnucl_spont (rhog, pgas, cpars.mseed, cpars.sigdotN,
                                cpars.past, cpars.sigast, cpars.gz)

    #mass flux conservation!
    Mv = -Mc

    #gradients
    #vsettl is positively defined
    Mn_z = rhodotN  -cpars.mseed *dndt_coag
    Mc_z = rhodotC
    xn_z = (-vsettl*xn -Mn/rhog) /cpars.Kzz
    xc_z = (-vsettl*xc -Mc/rhog) /cpars.Kzz
    xv_z = (           -Mv/rhog) /cpars.Kzz

    #reshape the bunch and re-normalize
    dum = npy.array([Mn_z, Mc_z, xn_z, xc_z, xv_z])
    out = (dum.T/cfg.normArr).T
    return out


def iterate_solution (xn, Yn, ttol, xnodes):
    """
    improve the solution based on a guess
    """

    cfg.ztop = x_to_z(xn.max())
    bvpsol = sciint.solve_bvp(Y_x, boundary_conditions, xn, Yn, tol=ttol, max_nodes=xnodes)

    return bvpsol


def normalize_norms (Yn, bvpsol, iphase):
    dumarr = npy.ones((cfg.Ndim))
    for k in range(cfg.Ndim):
        if cfg.inorm[k]!=0:
            dumarr[k] = (cfg.inorm[k]*bvpsol.y[k]).max()
        
    if (dumarr.min()<0.5 or dumarr.max()>1.5) and iphase>0:
        print('[cloud-model]:renormalizing', dumarr.min(), dumarr.max())
        renormalize = True
        for k in range(cfg.Ndim):
            if cfg.inorm[k]!=0:
                #limit the max change to fdum
                fdum = 1.1 -10*npy.log(cfg.ffac)
                dum = min(max(1/fdum,dumarr[k]),fdum)
                Yn[k,:] /= dum#already prepared above
                cfg.normArr[k] *= dum
    else:
        renormalize = False

    return renormalize
